#lang scheme
(require redex)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; lambda-v++ with levels and aspects
;
; version 1.1 - corresponds to AOSD.09 submission
; last updated Oct 25th, 2011
; by Éric Tanter
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;
; differences with the paper semantics:
; - supports a more full-fledged language
; - all lambda expressions are evaluated to lambda values 
;   that are *tagged* with a reference id (used for equality tests in pointcuts)
;   and a level (-1 for all functions, except level-capturing functions, tagged
;   with the level at which they always run
;;

(define-language λal
  ;; expressions
  (e (e e ...) (if e e e) x v
     (λ (x ...) e)
     (λs (x ...) e) ;; level-capturing lambda
     (let ((x e) ...) e) ; for convenience
     (begin e e ...)     ; .
     (up e) (down e) ; level shifting
     
     ;;---below: not user visible syntax ---;;
     (in-up e) (in-down e) ; 
     (in-shift l e)        ; 
     (jp j) (in-jp e)      ;
     (app/prim e e ...))
  
  ;; values and primitives
  (v (λ r l (x ...) e)    ;; each lambda is tagged with an id and its level of app (-1 for dyn)
     (list v ...) number #t #f 
     string prim unspecified)
  (prim  +  - list cons car cdr empty? eq? deploy)
  (r (variable-prefix r)) ;; to allow ref equality of functions
  
  ;; program context <l,J,A,E>
  (P (l J A E))
  (l number) ;; level
  
  ;; jp stack, jp, jp kind
  (J (stack j ...))
  (j (l k v v ...)) ;; kind, lambda, args
  (k "call") ;; here, only call
  
  ;; aspect environment 
  (A (asps a ...))
  (a (l v v)) ;; level, pc, adv
  
  ;; expr evaluation context
  (E (v ... E e ...)
     (if E e e)
     (begin E e e ...)
     (in-up E) (in-down E) (in-jp E) (in-shift l E)
     (app/prim v ... E e ...)
     hole)
  (x variable-not-otherwise-mentioned))


(define red
  (reduction-relation 
   λal   
   ;; function tagging
   (--> (in-hole P (λ (x ...) e))
        (in-hole P (λ ,(gensym "r") -1  (x ...) e))
        "FunTag")
   
   ; level capturing case
   (--> (l_1 J A (in-hole E (λs (x ...) e)))
        (l_1 J A (in-hole E (λ ,(gensym "r") l_1 (x ...) e)))
        "FunTag-c")
   
   ;; arithmetic
   (--> (in-hole P (+ number_1 number_2))
        (in-hole P ,(+ (term number_1) (term number_2)))
        "+")
   (--> (in-hole P (- number_1 number_2))
        (in-hole P ,(- (term number_1) (term number_2)))
        "-")
   
   ;; if
   (--> (in-hole P (if #f e_1 e_2))
        (in-hole P e_2)
        "if-f")
   (--> (in-hole P (if v_1 e_1 e_2))
        (in-hole P e_1)
        "if-t"
        (side-condition (term v_1)))  
   
   ;; let (syntactic sugar)
   (--> (in-hole P (let ((x e) ...) e_0))
        (in-hole P (app/prim (λ (x ...) e_0) e ...))
        "let")
   
   ;; begin
   (--> (in-hole P (begin v e_1 e_2 ...))
        (in-hole P (begin e_1 e_2 ...))
        "seq")
   (--> (in-hole P (begin e))
        (in-hole P e)
        "outseq") 
   
   ;; eq?
   (--> (in-hole P (eq? v_1 v_2))
        (in-hole P #t)
        "eq-t"
        (side-condition (term (feq? v_1 v_2))))
   (--> (in-hole P (eq? v_1 v_2))
        (in-hole P #f)
        "eq-f"
        (side-condition (not (term (feq? v_1 v_2))))) 
   
   ;; lists
   (--> (in-hole P (cons v_0 (list v_1 ...)))
        (in-hole P (list v_0 v_1 ...))
        "cons")
   (--> (in-hole P (car (list v_1 v_2 ...)))
        (in-hole P v_1)
        "car")
   (--> (in-hole P (cdr (list v_1 v_2 ...)))
        (in-hole P (list v_2 ...))
        "cdr")
   (--> (in-hole P (empty? (list)))
        (in-hole P #t)
        "empty?-t")
   (--> (in-hole P (empty? v_1))
        (in-hole P #f)
        "empty?-f"
        (side-condition (not (equal? (term v_1) (term (list))))))
   
   ;; shifting execution levels
   (--> (l_1 J A (in-hole E (up e)))  
        (,(add1 (term l_1)) J A (in-hole E (in-up e)))
        "InUp")
   (--> (l_1 J A (in-hole E (in-up v)))
        (,(sub1 (term l_1)) J A (in-hole E v))
        "OutUp")
   (--> (l_1 J A (in-hole E (down e)))
        (,(sub1 (term l_1)) J A (in-hole E (in-down e)))
        "InDwn")
   (--> (l_1 J A (in-hole E (in-down v)))
        (,(add1 (term l_1)) J A (in-hole E v))
        "OutDwn")
   
   ;; aspects
   ; deploy
   (--> (l_1 J (asps a ...) (in-hole E (deploy v_1 v_2)))
        (l_1 J (asps (,(add1 (term l_1)) v_1 v_2) a ...) (in-hole E unspecified))
        "Deploy")
   
   ; app/prim: beta red
   (--> (in-hole P (app/prim (λ r l (x ...) e) v ...))
        (in-hole P (subst-n (x v) ... e))
        "AppPrim"
        (side-condition (eq? -1 (term l))))
   
   ; app/prim for level-capturing function
   (--> (l_1 J A (in-hole E (app/prim (λ r l_2 (x ...) e) v ...)))
        (l_2 J A (in-hole E (in-shift l_1 (subst-n (x v) ... e))))
        "AppShift"
        (side-condition (not (eq? -1 (term l_2)))))
   
   ; back from shift
   (--> (l_2 J A (in-hole E (in-shift l_1 v)))
        (l_1 J A (in-hole E v))
        "Shift")
   
   ; app: call jp
   (--> (l_1 J A (in-hole E ((λ r l (x ...) e) v ...)))
        (l_1 J A (in-hole E (jp (,(add1 (term l_1)) "call" (λ r l (x ...) e) v ...))))
        "App")
   
   ; weaving:
   (--> (l_1 (stack j ...) (asps a ...) (in-hole E (jp (l_2 k (λ r l (x ...) e) v ...))))
        (l_1 (stack (l_2 k (λ r l (x ...) e) v ...) j ...) (asps a ...) 
             (in-hole E (in-jp (up (app/prim (W (l_2 k (λ r l (x ...) e) v ...)
                                                (asps a ...)) 
                                             v ...)))))
        "Weave")
   
   (--> (l_1 (stack (l_2 k v_0 v ...) j ...) A (in-hole E (in-jp v_1)))
        (l_1 (stack j ...) A (in-hole E v_1))
        "OutJp")
   ))

(define-metafunction λal
  W : j A -> e
  ;; generate woven advice, recurring on aspect environment
  ((W (l_1 k (λ r l (x ...) e) v ...) (asps (l_2 v_1 v_2) a ...))
   (app/prim (λ (p) ;;proceed
               (if (eq? l_1 l_2) ;; level check 
                   (let ((c (v_1 (list k (λ r l (x ...) e) v ...))))
                     (if c 
                         (λ (x ...) (v_2 p c x ...))
                         p))
                   p))
             (W (l_1 k (λ r l (x ...) e) v ...) (asps a ...))))  ;; base case
  
  ;; empty aspect environment (base case) -> just proceed at level-1
  ((W (l_1 k (λ r l (x ...) e) v ...) (asps))
   ;; we capture the level of original computation, which was already incremented when
   ;; emiting the join point, so we use one level below that of the join point
   (λ ,(gensym "r") ,(sub1 (term l_1)) (x ...) (app/prim (λ r l (x ...) e) x ...))))


;; comparison function: uses eq? on all terms, except for 
;; λ values, uses reference equality
(define-metafunction λal
  ((feq? (λ r_1 l_1 (x_1 ...) e_1) (λ r_1 l_2 (x_2 ...) e_2))
   #t)
  ((feq? (λ r_1 l_1 (x_1 ...) e_1) (λ r_2 l_2 (x_2 ...) e_2))
   #f
   (side-condition (not (eq? (term r_1) (term r_2)))))
  ((feq? any_1 any_2)
   ,(eq? (term any_1) (term any_2))))


;;;; substitution functions
(define-metafunction λal
  subst-n : (x any) ... any -> any 
  [(subst-n (x_1 any_1) (x_2 any_2) ... any_3) 
   (subst x_1 any_1 (subst-n (x_2 any_2) ... any_3))] 
  [(subst-n any_3) any_3])

(define-metafunction λal 
  subst : x any any -> any  
  ;; 1. x_1 bound, so don't continue in λ body  
  [(subst x_1 any_1 (λ r l (x_2 ... x_1 x_3 ...) any_2)) 
   (λ r l (x_2 ... x_1 x_3 ...) any_2) 
   (side-condition (not (member (term x_1) (term (x_2 ...)))))]  
  
  ;; 2. general purpose capture avoiding case  
  [(subst x_1 any_1 (λ r l (x_2 ...) any_2)) 
   ,(term-let ([(x_new ...) 
                (variables-not-in (term (x_1 any_1 any_2)) 
                                  (term (x_2 ...)))]) 
              (term 
               (λ r l (x_new ...) 
                 (subst x_1 any_1 (subst-vars (x_2 x_new) ... any_2)))))]  
  
  ;; repeated, for substituting in lambdas that are not values yet
  
  ;; 1'. x_1 bound, so don't continue in λ body  
  [(subst x_1 any_1 (λ (x_2 ... x_1 x_3 ...) any_2)) 
   (λ (x_2 ... x_1 x_3 ...) any_2) 
   (side-condition (not (member (term x_1) (term (x_2 ...)))))]  
  
  ;; 2'. general purpose capture avoiding case  
  [(subst x_1 any_1 (λ (x_2 ...) any_2)) 
   ,(term-let ([(x_new ...) 
                (variables-not-in (term (x_1 any_1 any_2)) 
                                  (term (x_2 ...)))]) 
              (term 
               (λ (x_new ...) 
                 (subst x_1 any_1 (subst-vars (x_2 x_new) ... any_2)))))]    
  
  ;; same as above for level capturing lambdas that are not values yet
  ;; 1'. x_1 bound, so don't continue in λ body  
  [(subst x_1 any_1 (λs (x_2 ... x_1 x_3 ...) any_2)) 
   (λs (x_2 ... x_1 x_3 ...) any_2) 
   (side-condition (not (member (term x_1) (term (x_2 ...)))))]  
  
  ;; 2'. general purpose capture avoiding case  
  [(subst x_1 any_1 (λs (x_2 ...) any_2)) 
   ,(term-let ([(x_new ...) 
                (variables-not-in (term (x_1 any_1 any_2)) 
                                  (term (x_2 ...)))]) 
              (term 
               (λs (x_new ...) 
                   (subst x_1 any_1 (subst-vars (x_2 x_new) ... any_2)))))]
  
  ;; 3. replace x_1 with e_1  
  [(subst x_1 any_1 x_1) any_1]  
  ;; 4. x_1 and x_2 are different, so don't replace  
  [(subst x_1 any_1 x_2) x_2]  
  ;; the last two cases cover all other expression forms  
  [(subst x_1 any_1 (any_2 ...)) 
   ((subst x_1 any_1 any_2) ...)] 
  [(subst x_1 any_1 any_2) any_2])

(define-metafunction λal 
  subst-vars : (x any) ... any -> any
  [(subst-vars (x_1 any_1) x_1) any_1] 
  [(subst-vars (x_1 any_1) (any_2 ...)) ((subst-vars (x_1 any_1) any_2) ...)] 
  [(subst-vars (x_1 any_1) any_2) any_2] 
  [(subst-vars (x_1 any_1) (x_2 any_2) ... any_3) 
   (subst-vars (x_1 any_1) (subst-vars (x_2 any_2) ... any_3))] 
  [(subst-vars any) any])

;;;;;;;;;;;;;;;;;;;;;;;; TEST SUITE ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; Testing let
(test-->>
 red
 (term 
  (0 (stack) (asps)
     (let ((x 3) (y (+ 2 2))) (+ x y))))
 (term (0 (stack) (asps) 7)))

;; Testing lists and if

(test-->>
 red
 (term
  (0 (stack) (asps)
     (car (cdr (cons 3 (list 1 2))))))
 (term (0 (stack) (asps) 1)))

(test-->>
 ; (traces
 red
 (term
  (0 (stack) (asps)
     (empty? (list))))
 ;)
 (term (0 (stack) (asps) #t)))

(test-->>
 red
 (term
  (0 (stack) (asps)
     (empty? (list 1 3))))
 (term (0 (stack) (asps) #f)))

(test-->>
 red
 (term
  (0 (stack) (asps)
     (empty? (cdr (cdr (cons 2 (cons 1 (list))))))))
 (term (0 (stack) (asps) #t)))

(test-->>
 red
 (term
  (0 (stack) (asps)
     (if (empty? (cdr (cdr (cons 2 (cons 1 (list))))))
         1
         2)))
 (term (0 (stack) (asps) 1)))

(test-->>
 red
 (term
  (0 (stack) (asps)
     (if (empty? (cdr (cons 2 (list 1))))
         1
         2)))
 (term (0 (stack) (asps) 2)))

(test-->>
 red
 (term
  (0 (stack) (asps)
     (if (cdr (cdr (list 1 2)))
         1
         2)))
 (term (0 (stack) (asps) 1)))

;; Testing eq?
(test-->>
 red
 (term 
  (0 (stack) (asps)
     (if (eq? (+ 2 3) 0)
         1
         0)))
 (term (0 (stack) (asps) 0)))


(test-->>
 red
 (term 
  (0 (stack) (asps)
     (if (eq? (+ 2 3) (+ 1 4))
         1
         0)))
 (term (0 (stack) (asps) 1)))

(test-->>
 red
 (term 
  (0 (stack) (asps)
     (if (eq? "hola" "hola")
         1
         0)))
 (term (0 (stack) (asps) 1)))

; function identity check
(test-->>
 red 
 (term
  (0 (stack) (asps)
     (let ((f (λ (x) x)))
       (eq? f f))))
 (term (0 (stack) (asps) #t)))

(test-->>
 ;(traces 
 red
 (term
  (0 (stack) (asps)
     ((λ (f)
        (if (eq? (f 4) 5)
            ((λ () #t))
            ((λ () #f))))
      ((λ (x) 
         (λ (y) (+ x y))) 2))
     ))
 ;)
 (term (0 (stack) (asps) #f)))

;; Testing aspects
(test-->>
 ; (traces 
 red
 (term
  (0 (stack) (asps (0 (λ r1 -1 (jpt) (list))
                      (λ r2 -1 (q c x y) (q x y))))
     ((λ (x y) (+ x y))
      2 3)))
 ;)
 (term (0 (stack) (asps (0 (λ r1 -1 (jpt) (list))
                           (λ r2 -1 (q c x y) (q x y)))) 5)))

(test-->>
; (traces 
 red
 (term
  ;; aspect level changed to 1
  (0 (stack) (asps (1 (λ r1 -1 (jpt) (list))
                      (λ r2 -1 (q c x y) (- x y))))
     ((λ (x y) (+ x y))
      2 3)))
 ;)
 (term (0 (stack) (asps (1 (λ r1 -1 (jpt) (list))
                           (λ r2 -1 (q c x y) (- x y)))) 
          -1)))

;; 2 aspects, each changing an arg
(test-->>
; (traces 
 red
 (term
  (0 (stack) (asps) 
     (begin (deploy (λ r1 -1 (jpt) (list 10))
                    (λ r2 -1 (q c x y) (q (car c) y)))
            (deploy (λ r1 -1 (jpt) (list 1))
                    (λ r2 -1 (q c x y) (q x (car c))))
            
            ((λ (x y) (+ x y))
             2 3))))
 ;)
 ;;aspects level changed to 1
 (term (0 (stack) (asps (1 (λ r1 -1 (jpt) (list 1))
                           (λ r2 -1 (q c x y) (q x (car c))))
                        (1 (λ r1 -1 (jpt) (list 10))
                           (λ r2 -1 (q c x y) (q (car c) y)))) 
          11)))

;; 1 aspect on each level (2)
(test-predicate
 (λ (res)
   (redex-match λal (0 (stack) A_1 7)
                (car res)))  
 (apply-reduction-relation* 
  red
  (term
   (0 (stack) (asps) 
      (let ((pc0 (λ (jpt) (list 1))))
        (begin 
          ;; level 1 -> replaces binding returned by pointcut of level 0 aspect
          (up (deploy (λ (jpt) (if (eq? (car (cdr jpt)) pc0)
                                   (list)
                                   #f))
                      (λ (q c x) (list 5))))
          
          ;; level 0 -> replaces 2nd parameter of base call
          (deploy pc0
                  (λ (q c x y) (q x (car c))))
          
          ((λ (x y) (+ x y))
           2 3)))))))

;; 1 aspect on each level (3 / same as above + level2 aspect overriding advice)
;(test-->>
(test-predicate
 (λ (res)
   (redex-match λal (0 (stack) A_1 12)
                (car res)))
 (apply-reduction-relation*
  red
  (term
   (0 (stack) (asps)
      (let ((adv1 (λ (q c x) (list 5)))
            (pc0  (λ (jpt) (list 1))))
        (begin 
          ;; level 2 -> replaces value returned by advice of level 1 aspect
          (up (up (deploy (λ (jpt) (if (eq? (car (cdr jpt)) adv1)
                                       (list)
                                       #f))
                          (λ (p c bp bc bx) (list 10)))))
          
          ;; level 1 -> replaces binding returned by pointcut of level 0 aspect
          (up (deploy (λ (jpt) (if (eq? (car (cdr jpt)) pc0) 
                                   (list)
                                   #f))
                      adv1))
          
          ;; level 0 -> replaces 2nd parameter of base call
          (deploy pc0 (λ (q c x y) (q x (car c))))
          
          ((λ (x y) (+ x y))
           2 3)))))))

;; level-capturing functions
(test-predicate
 (λ (res)
   (redex-match λal (0 (stack) A_1 4)
                (car res)))
 (apply-reduction-relation*
  red
  (term (0 (stack) (asps)
           (let ((f (λ (x) (+ x x)))
                 (task (up (λs (x) (f x))))) ;; task runs at level 1
             (begin 
               (deploy (λ (jpt) (eq? (car (cdr jpt)) f)) ;; should not see f apply
                       (λ (q c x) 5))
               (task 2)))))))

(test-predicate
 (λ (res)
   (redex-match λal (0 (stack) A_1 5)
                (car res)))
 (apply-reduction-relation*
  red
  (term (0 (stack) (asps)
           (let ((f (λ (x) (+ x x)))
                 (task (up (λs (x) (f x))))) ;; runs at level 1
             (begin 
               (up (deploy (λ (jpt) (eq? (car (cdr jpt)) f)) ;; up: sees f apply
                           (λ (q c x) 5)))
               (task 2)))))))


(test-results)